/*
 (c) James Gregson (james.gregson@gmail.com), 2013
 Free for academic and personal use with proper attribution, do not redistribute,
 use at your own risk.
 
 Sample (and simple) implementation of Stochastic Deconvolution, from the
 CVPR 2013 poster of the same name.  This is NOT the code used in the paper,
 it is only intended to illustrate the algorithm and provide other
 researchers/hobbyists with a starting point for experimenting with the method.
 
 The stochastic sampling algorithm. This may differ slightly from the exact
 pseudocode in the paper, I have experimented a bit with the acceptance
 criteria for mutations.
*/
#include<cstdlib>
#include<algorithm>
#include"stochastic_deconvolution.h"

double stochastic_devonvolution( sd_data *data, sd_callbacks *cb, double ed, int n_mutations ){
	double fx, fy, a_rate = 0.0f;
    sd_sample sample_x, sample_y;
    
    // initialize the first sample randomly within the image
    sample_x.x = drand48()*data->input.width();
    sample_x.y = drand48()*data->input.height();
    sample_x.ed = ed;
    
	// evaluate the first sample
	fx = cb->evaluate( data, &sample_x );
	
	// loop until it's time to stop
	for( int i=0; i<n_mutations; i++ ){
		
		// generate a mutation and evaluate the objective function
		cb->mutate( data, &sample_x, &sample_y );
		fy = cb->evaluate( data, &sample_y );
		
		// if the mutation helps, apply it
		if( fy > 0.0 ){
			a_rate += 1.0f;
            cb->splat( data, &sample_y, 1.0 );
		}
		
        // if x worsens the objective, and y worsens it by less, then
        // accept y. otherwise use the standard MH acceptance criteria.
        // Doesn't handle mixed-signs of fx and fy properly, but it doesn't
        // seem to harm the algorithm
		if( (fx <= 0.0 && fy >= fx) || (drand48() < std::min( 1.0, fy/fx ) ) ){
			cb->copy( data, &sample_y, &sample_x );
			fy = cb->evaluate( data, &sample_x );
			fx = fy;
		}
	}
    
    // return the fraction of accepted samples in order to schedule
    // decreases to ed
	return a_rate/double(n_mutations);
}